# coding=utf-8
""" Contains objects related to points.

"""

from shapely.geometry import Point as ShapelyPoint
from shapely.wkt import loads as wkt_to_geom
from src.oogeo.objects.geometrybase import GeometryBase
import src.oogeo.objects.polygon


class Point(GeometryBase):
    """ Provides support for analyzing point features in GIS layers.

    """

    def __init__(self, geometry, spatial_reference, allow_invalid=False):
        """ Initializes the point object.

        :param geometry: The geometry object representing a point.
        """
        # use shapely to construct a wkt string for creating new ogr geometry objects
        # (required to work around a bug in ogr)
        if isinstance(geometry, str):
            geometry = wkt_to_geom(geometry)
        elif isinstance(geometry, tuple):
            geometry = ShapelyPoint(geometry[:2])
        elif isinstance(geometry, dict):
            geometry = ShapelyPoint((geometry['x'], geometry['y']))

        assert isinstance(geometry, ShapelyPoint), 'A point must be provided'

        super(Point, self).__init__(geometry, spatial_reference, allow_invalid)

    @property
    def x(self):
        """ The 'x' coordinate for the point.

        :return: A float value representing the 'x' portion of the coordinates.
        """
        return self.geometry.x

    @property
    def y(self):
        """ The 'y' coordinate for the point.

        :return: A float value representing the 'y' portion of the coordinates.
        """
        return self.geometry.y

    def buffer(self, distance):
        """ Creates a circular buffer around the point at the specified distance.

        :param distance: The buffer distance.
        :return: A `Polygon()` object representing the buffer around the point
        """
        return src.oogeo.objects.polygon.Polygon(self.geometry.buffer(distance), self.spatial_reference)

    def snap_to_vertex(self, other_geometry, tolerance, allow_invalid=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """
        return Point(self._snap_(other_geometry, tolerance), self.spatial_reference, allow_invalid=allow_invalid)

    def snap_to_edge(self, other_geometry, tolerance, allow_invalid=False, densify=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """

        snap_point = other_geometry.snap_point(self) if not isinstance(other_geometry, Point) else other_geometry
        return_geometry = self.snap_to_vertex(snap_point, tolerance=tolerance, allow_invalid=allow_invalid)

        return return_geometry

    def clone(self):
        """ Creates a copy of the geometry.

        :return: A copy of the geometry.
        """
        return Point(ShapelyPoint(self.geometry), self.spatial_reference, allow_invalid=self.allow_invalid)

    def scale(self, scale_factor):
        """ Scales the shape by multiplying each of it's vertices by the provided scale factor.

        :param scale_factor: The factor to multiple each x/y value by for each vertex in the shape.
        :return: A new geometry object with the x/y values scaled by the provided scale factor.
        """
        return Point({'x': self.x * scale_factor, 'y': self.y * scale_factor}, self.spatial_reference)

    def transform(self, delta_x, delta_y):
        """ Moves the geometry by the specified delta values by adding the delta values to the x/y
        values of each vertex in the geometry. Providing positive delta values will generally move
        the geometry to the north and east, whereas negative values will generally move the geometry
        to the south and west.

        :param delta_x: The value to add to the x coordinate of each vertex in the geometry.
        :param delta_y: The value to add to the y coordinate of each vertex in the geometry.
        :return: A new geometry that has been transformed based on the provided inputs.
        """
        return Point({'x': self.x + delta_x, 'y': self.y + delta_y}, self.spatial_reference)



"""
Unit tests for the Table class. For more information see the
[API Tests](https://github.com/envelopecity/GIS/tree/dev#api-tests) section of the GIS repository.
"""

import unittest

from src.oogeo.unittests.databasetest import DatabaseTest
from src.oogeo.objects.table import Table
from src.oogeo.objects.workspace import Workspace


class TableClassUnitTests(DatabaseTest):

    def setUp(self):
        # load the test table
        if not hasattr(self, 'test_table'):
            self.workspace = Workspace(self._databases['testing_inputs'])
            self.test_table = Table('adjacent_lots', self.workspace)

    def test_create_table_from_workspace(self):
        # verify that tables created by the workspace object function correctly
        with self.workspace.create_table('create_table_test', schema='testing') as table:
            table.add_field('test_str', 'TEXT')
            table.add_field('test_int', 'SHORT')
            table.add_field('test_dbl', 'DOUBLE')
            table.add_field('test_bool', 'BOOLEAN')
            rows = [
                {
                    'test_str': 'A',
                    'test_int': 1,
                    'test_dbl': 1.1,
                    'test_bool': True,
                },
                {
                    'test_str': 'B',
                    'test_int': 2,
                    'test_dbl': 2.2,
                    'test_bool': False,
                },
            ]
            table.add_rows(rows)
            self.assertEqual(table.count, 2)

            table.select_layer_by_attribute('test_int = 1')
            with table.get_search_cursor(['test_str', 'test_dbl', 'test_bool']) as cursor:
                row = next(cursor)
            self.assertEqual(row['test_str'], 'A')
            self.assertEqual(row['test_dbl'], 1.1)
            self.assertTrue(row['test_bool'])

            table.select_layer_by_attribute('test_int = 2')
            with table.get_search_cursor(['test_str', 'test_dbl', 'test_bool']) as cursor:
                row = next(cursor)
            self.assertEqual(row['test_str'], 'B')
            self.assertEqual(row['test_dbl'], 2.2)
            self.assertFalse(row['test_bool'])
        self.workspace.drop_schema('testing')

    def test_initialize_layer(self):
        # verify we can initialize a layer object without failure
        self.assertIsInstance(self.test_table, Table)

    def test_select_layer_by_attribute(self):
        # get the total rows in the table
        with self.test_table.copy_to_new_table(self.test_table.table_name, output_schema='testing') as local_table:
            total_rows = local_table.count

            # make a selection in the layer and test it
            local_table.select_layer_by_attribute('OBJECTID = 1')
            self.assertEqual(local_table.get_selection_count(), 1)
            self.assertEqual(local_table.count, 1)

            # select two more to test the add to selection function
            local_table.select_layer_by_attribute('OBJECTID = 2 OR OBJECTID = 3', selection_type='ADD_TO_SELECTION')
            self.assertEqual(local_table.get_selection_count(), 3)
            self.assertEqual(local_table.count, 3)

            # remove one from the selection set with the subset selection option
            local_table.select_layer_by_attribute('OBJECTID = 2 OR OBJECTID = 3', selection_type='SUBSET_SELECTION')
            self.assertEqual(local_table.get_selection_count(), 2)
            self.assertEqual(local_table.count, 2)

            # remove one from the selection set with the remove from option
            local_table.select_layer_by_attribute('OBJECTID = 2', selection_type='REMOVE_FROM_SELECTION')
            self.assertEqual(local_table.get_selection_count(), 1)
            self.assertEqual(local_table.count, 1)

            # switch the selection
            local_table.select_layer_by_attribute('OBJECTID = 2', selection_type='SWITCH_SELECTION')
            self.assertEqual(local_table.get_selection_count(), total_rows - 1)
            self.assertEqual(local_table.count, total_rows - 1)

            local_table.select_layer_by_attribute('OBJECTID = 1')
            self.assertEqual(local_table.get_selection_count(), 1)
            self.assertEqual(local_table.count, 1)

            # clear the selection
            local_table.clear_selection()
            self.assertEqual(local_table.get_selection_count(), total_rows)
            self.assertEqual(local_table.count, total_rows)

        # clean up testing schema
        self.workspace.drop_schema('testing')

    def test_oid_property(self):

        # test the oid property
        self.assertEqual(self.test_table.oid, 'OBJECTID')

    def test_max_oid(self):
        self.assertEqual(self.test_table.max_oid, 5)

    def test_field_names_property(self):
        self.assertIn('CBBL', self.test_table.field_names)
        self.assertIn('nbr_CBBL', self.test_table.field_names)
        self.assertIn('NODE_COUNT', self.test_table.field_names)
        self.assertIn('LENGTH', self.test_table.field_names)
        self.assertNotIn('OBJECTID', self.test_table.field_names)
        self.assertNotIn('SHAPE', self.test_table.field_names)

    def test_duplicate_table(self):
        with self.test_table.duplicate_table('duplicated_table_test', output_schema='testing') as duplicate_table:
            self.assertIsInstance(duplicate_table, Table)
            self.assertEqual(duplicate_table.count, 0)
            for field in self.test_table.field_names:
                self.assertIn(field, duplicate_table.field_names)
        self.workspace.drop_schema('testing')

    def test_fields_property(self):

        # test the field property
        fields = [field.name for field in self.test_table.fields]
        self.assertIn('CBBL', fields)

    def test_switch_and_clear_selection(self):

        # get the total rows in the table
        total_rows = self.test_table.count

        # make a selection in the layer, switch the selection, and test it
        self.test_table.select_layer_by_attribute('OBJECTID = 1')
        self.test_table.switch_selection()
        self.assertEqual(self.test_table.count, total_rows - 1)

        # clear the selection and test that
        self.test_table.clear_selection()
        self.assertEqual(self.test_table.count, total_rows)

    def test_field_functions(self):
        # add a new field to the table
        with self.test_table.copy_to_new_table('field_test', output_schema='testing') as test_table:
            test_table.add_field('test', 'TEXT')

            # test that the new field is in the list of fields
            fields = test_table.field_names
            self.assertIn('test', fields)

            # calculate a value into the field
            test_table.calculate_field('test', "'Test Text'")

            # add a second new field with a precalculated value
            test_table.add_field('TEST_2', 'TEXT',
                default_value="'Test Text'")

            # select a row in the table and verify calculated field
            with test_table.get_search_cursor(['test', 'TEST_2'],
                query_string='OBJECTID = 1') as cursor:
                for row in cursor:
                    self.assertEqual(row['test'], 'Test Text')
                    self.assertEqual(row['TEST_2'], 'Test Text')

            # delete one of the columns and verify
            test_table.delete_field('test')

            # rename the test_2 field to test_3
            test_table.rename_field('TEST_2', 'TEST_3')
            fields = test_table.field_names
            self.assertIn('TEST_3', fields)
            self.assertNotIn('TEST_2', fields)

            # add a new field and test the copy field values function
            test_table.add_field('test_4', 'TEXT')
            test_table.copy_field_values('TEST_3', 'test_4')

            # add a new field and test copying with the calculate field function
            test_table.add_field('test_5', 'TEXT')
            test_table.add_field('test_6', 'LONG')
            test_table.add_field('test_7', 'LONG')
            test_table.calculate_field('test_5', '!test_4!')
            test_table.calculate_field('test_6', '!{}!'.format(test_table.oid))
            test_table.calculate_field('test_7', '!test_6! * 2')
            # verify that the values in test 3 were copied to test 4
            with test_table.get_search_cursor(['TEST_3', 'test_4', 'test_5', 'test_6', 'test_7'],
                query_string='OBJECTID = 1') as cursor:
                for row in cursor:
                    self.assertEqual(row['TEST_3'], row['test_4'])
                    self.assertEqual(row['test_4'], row['test_5'])
                    self.assertEqual(row['test_6'], 1)
                    self.assertEqual(row['test_7'], 2)

            # delete the other field indirectly using the delete other fields method and test
            test_table.delete_other_fields(['nbr_CBBL', 'LENGTH', 'NODE_COUNT', 'CBBL'])

            # test that the new field is not in the list of fields
            fields = test_table.field_names
            self.assertIn('nbr_CBBL', fields)
            self.assertIn('LENGTH', fields)
            self.assertIn('NODE_COUNT', fields)
            self.assertIn('CBBL', fields)
            self.assertNotIn('TEST_3', fields)
            self.assertNotIn('test_4', fields)
            self.assertNotIn('test_5', fields)
            self.assertNotIn('test_6', fields)
        self.workspace.drop_schema('testing')

    def test_get_field_type(self):
        # verify the field types in the test table
        self.assertEqual(self.test_table.get_field_type('CBBL'), 'VARCHAR(255)')
        self.assertEqual(self.test_table.get_field_type('NODE_COUNT'), 'INTEGER')
        self.assertEqual(self.test_table.get_field_type('LENGTH'), 'DOUBLE PRECISION')

    def test_get_unique_field_values(self):
        # get the list of distinct node_count values and test it (there is only 1 value in the table - 0)
        node_counts = self.test_table.get_unique_field_values('NODE_COUNT')
        self.assertEqual(len(node_counts), 1)
        for node_count in node_counts:
            self.assertEqual(node_count, 0)
        cbbls = self.test_table.get_unique_field_values('CBBL')
        self.assertEqual(len(cbbls), 4)
        self.assertIn('NY-MN-10-15', cbbls)
        self.assertIn('NY-MN-10-16', cbbls)
        self.assertIn('NY-MN-10-1', cbbls)
        self.assertIn('NY-MN-10-14', cbbls)

    def test_get_distinct_field_values(self):
        single_field = self.test_table.get_distinct_field_values('NODE_COUNT')
        self.assertEqual(len(single_field), 1)
        self.assertIn({'NODE_COUNT': 0}, single_field)

        multi_fields = self.test_table.get_distinct_field_values(['NODE_COUNT', 'CBBL'])
        self.assertEqual(len(multi_fields), 4)
        self.assertIn({'NODE_COUNT': 0, 'CBBL': 'NY-MN-10-15'}, multi_fields)
        self.assertIn({'NODE_COUNT': 0, 'CBBL': 'NY-MN-10-16'}, multi_fields)
        self.assertIn({'NODE_COUNT': 0, 'CBBL': 'NY-MN-10-1'}, multi_fields)
        self.assertIn({'NODE_COUNT': 0, 'CBBL': 'NY-MN-10-14'}, multi_fields)

    def test_row_manipulation_methods(self):
        with self.test_table.copy_to_new_table(self.test_table.table_name, output_schema='testing') as local_table:
            # get a baseline count of the number of rows in the table
            base_count = local_table.count

            # insert a row into the table
            new_rows = [
                {'CBBL': 'test-cbbl-value', 'nbr_CBBL': 'test-nbr-cbbl-value', 'LENGTH': 0.1, 'NODE_COUNT': 0}
            ]
            local_table.add_rows(new_rows)

            # verify that the new row has been added
            self.assertEqual(local_table.count, base_count + 1)

            # get the new row with an update cursor and update it (the new row will have an oid > 5)
            with local_table.get_update_cursor('NODE_COUNT', 'OBJECTID > 5') as cursor:
                for row in cursor:
                    row['NODE_COUNT'] = 1
                    cursor.update_row(row)

            # query the new row back out with a search cursor and verify the update
            with local_table.get_search_cursor('NODE_COUNT', 'OBJECTID > 5') as cursor:
                for row in cursor:
                    self.assertEqual(row['NODE_COUNT'], 1)

            # delete the new row and verify that it's gone
            local_table.select_layer_by_attribute('OBJECTID > 5')
            local_table.delete_selected_rows()
            local_table.clear_selection()
            self.assertEqual(local_table.count, base_count)
        self.workspace.drop_schema('testing')

    def test_copy_to_new_table_and_delete_table(self):
        # select a few rows and copy them to a new table
        self.test_table.select_layer_by_attribute('OBJECTID > 3')
        copy_table = self.test_table.copy_to_new_table('copy_table_test')
        self.test_table.clear_selection()

        # verify the table has been added
        table_list = self.workspace.list_tables()
        self.assertEqual(len(table_list), 2)
        self.assertIn('copy_table_test', table_list)

        # verify the fields in the tables match
        base_fields = [field.name for field in self.test_table.fields]
        copy_fields = [field.name for field in copy_table.fields]
        self.assertListEqual(base_fields, copy_fields)

        # test adding an index to the table
        copy_table.add_index('CBBL')
        self.assertIn('idx_copy_table_test_CBBL', copy_table.list_indexes())

        # verify there are 2 rows in the new table
        self.assertEqual(copy_table.count, 2)

        # delete the new table and verify it's gone
        copy_table.delete()
        table_list = self.workspace.list_tables()
        self.assertEqual(len(table_list), 1)
        self.assertNotIn('copy_table', table_list)

    def test_append(self):
        # select a few rows and copy them to a new table
        self.test_table.select_layer_by_attribute('OBJECTID > 3')
        copy_table = self.test_table.copy_to_new_table('copy_table_test')
        self.test_table.clear_selection()

        # append the test table to the copy table and verify the row count
        combined_count = self.test_table.count + copy_table.count
        copy_table.append(self.test_table, 'TEST')
        self.assertEqual(copy_table.count, combined_count)

        # delete the new table
        copy_table.delete()

    def test_delete_duplicates(self):
        # get a baseline count of the number of rows in the table
        base_count = self.test_table.count

        # duplicate a row in the table
        new_rows = []
        with self.test_table.get_search_cursor(['CBBL', 'nbr_CBBL', 'LENGTH', 'NODE_COUNT'],
            query_string='OBJECTID = 1') as cursor:
            for row in cursor:
                # insert a row into the table
                new_rows.append({'CBBL': row['CBBL'], 'nbr_CBBL': row['nbr_CBBL'], 'LENGTH': row['LENGTH'], 'NODE_COUNT': row['NODE_COUNT']})
        self.test_table.add_rows(new_rows)
        # verify that the new row has been added
        self.assertEqual(self.test_table.count, base_count + 1)

        # delete duplicate rows
        self.test_table.delete_duplicates(['CBBL', 'nbr_CBBL', 'LENGTH', 'NODE_COUNT'])

        # verify that one row is gone
        self.assertEqual(self.test_table.count, base_count)

    def test_list_fields(self):
        fields = self.test_table.list_fields()
        self.assertEqual(len(fields), 5)
        fields = self.test_table.list_fields(type='String')
        self.assertEqual(len(fields), 2)
        fields = self.test_table.list_fields(type='Integer')
        self.assertEqual(len(fields), 2)
        fields = self.test_table.list_fields(type='Float')
        self.assertEqual(len(fields), 1)

    def test_delete_rows(self):
        # get a baseline count of the number of rows in the table
        base_count = self.test_table.count

        # duplicate a row in the table
        new_rows = []
        with self.test_table.get_search_cursor(['CBBL', 'nbr_CBBL', 'LENGTH', 'NODE_COUNT'],
            query_string='OBJECTID = 1') as cursor:
            for row in cursor:
                # insert a row into the table
                new_rows.append({'CBBL': row['CBBL'], 'nbr_CBBL': row['nbr_CBBL'], 'LENGTH': row['LENGTH'], 'NODE_COUNT': row['NODE_COUNT']})
        self.test_table.add_rows(new_rows)
        # verify that the new row has been added
        self.assertEqual(self.test_table.count, base_count + 1)

        # delete the new row
        self.test_table.select_layer_by_attribute('{} > 5'.format(self.test_table.oid))
        self.test_table.delete_selected_rows()

        # verify that one row is gone
        self.assertEqual(self.test_table.count, base_count)

    # def test_to_csv(self):
    #     # get a baseline count of the number of rows in the table
    #     base_count = self.test_table.count
    #
    #     # find the export path
    #     unittest_root = path.abspath(path.dirname(__file__))
    #     code_root = path.dirname(unittest_root)
    #     package_root = path.dirname(code_root)
    #     working_directory = path.join(package_root, self._configurations.get('WorkingDirectories', 'working_directory'))
    #
    #     # export the table to csv
    #     csv_fields = ['OBJECTID', 'CBBL', 'nbr_CBBL', 'LENGTH', 'NODE_COUNT']
    #     csv_file = path.join(working_directory, 'test_export.csv')
    #     self.test_table.to_csv(csv_file, csv_fields)
    #
    #     # read the exported csv file and make sure it's correct
    #     with open(csv_file, 'r') as file_reader:
    #         csv_reader = csv.reader(file_reader)
    #         csv_records = [
    #             {'OBJECTID': row[0], 'CBBL': row[1], 'nbr_CBBL': row[2], 'LENGTH': row[3], 'NODE_COUNT': row[4]}
    #             for row in csv_reader]
    #
    #     # verify that there is one more record in the csv file than in the table (this is the header row)
    #     self.assertEqual(len(csv_records), base_count + 1)
    #
    #     # verify that all of the fields are in the csv
    #     output_fields = sorted([value for value in csv_records[0]])
    #     csv_fields = sorted(csv_fields)
    #     self.assertListEqual(output_fields, csv_fields)
    #
    #     # delete the test file
    #     remove(csv_file)



if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TableClassUnitTests)
    unittest.TextTestRunner(verbosity=2).run(suite)



# coding=utf-8
""" Base class for all geometry objects. Contains properties and methods common to all geometry types.

"""

from shapely.geometry import Polygon as ShapelyPolygon, MultiPolygon, \
    Point as ShapelyPoint, LineString, MultiLineString, MultiPoint, GeometryCollection
from shapely.ops import nearest_points

from abc import ABC as abstract_base

import src.oogeo
from src.oogeo.objects.geometrybase import GeometryBase
from src.oogeo.functions.constants import BUFFER_END_STYLES, BUFFER_JOIN_STYLES
from src.oogeo.functions.conversions import string_to_units


def get_near_points(target_geometry, near_geometry, tolerance):
    """ Find the vertices in the target geometry that are close to the near geometry
    (within the specified tolerance).

    :param target_geometry: The geometry containing the vertices to be analyzed
    :param near_geometry: The geometry to measure to.
    :param tolerance: The maximum distance a point can be away from the near geometry
        and be considered a "near point"
    :return: A list of `Point()` objects representing the vertices of the target
        geometry that are near the near geometry.
    """

    # loop through the vertices of the target geometry and
    # find the points close to the near geometry
    return_list = []

    if isinstance(near_geometry.geometry, MultiPolygon):
        near_geometries = near_geometry.to_lines()
    elif isinstance(near_geometry.geometry, MultiLineString):
        near_geometries = near_geometry.split_line_at_vertices()
    else:
        near_geometries = [near_geometry]

    target_points = [target_geometry] if isinstance(target_geometry.geometry, ShapelyPoint) else target_geometry.points

    if isinstance(target_geometry.geometry, MultiPolygon):
        target_edges = target_geometry.to_lines()
    elif isinstance(target_geometry.geometry, MultiLineString):
        target_edges = target_geometry.split_line_at_vertices()
    else:
        target_edges = []

    for index, point in enumerate(target_points):
        distance = None
        edge = None
        for near_edge in near_geometries:
            test_distance = point.distance_to(near_edge)
            if test_distance > tolerance:
                continue
            if not distance or distance > test_distance:
                distance = test_distance
                edge = near_edge
        if not edge:
            continue

        # if we found a near edge do an out-front test
        nearest_point = edge.nearest_point(point) if isinstance(edge, src.oogeo.objects.line.Line) else edge
        if not nearest_point.equals(point):
            test_line = src.oogeo.objects.line.Line([point, nearest_point], point.spatial_reference)
            if not test_line.is_perpendicular(edge, tolerance=3):
                continue
            crosses_other_edge = False
            for target_edge in target_edges:
                if target_edge.first_point.equals(point) or target_edge.last_point.equals(point):
                    continue
                if test_line.crosses(target_edge):
                    crosses_other_edge = True
                    break
            if crosses_other_edge:
                continue

        # if we make it here the out-front test passed
        # now check for nearby vertices that are better candidates
        previous_point = target_points[index - 1] if index > 0 else None
        next_point = target_points[index + 1] if index + 1 < len(target_points) else None

        analyze = True
        for test_point in [previous_point, next_point]:
            if not test_point:
                continue

            # if the test point is further away from the edge than this point, don't consider it
            if test_point.distance_to(edge) > distance:
                continue

            # if the test point is closer to the edge than this point, do
            # and out-front test on the test point - if its not out front don't consider it
            test_nearest_point = edge.nearest_point(test_point)
            if not test_point.equals(test_nearest_point):
                test_line = src.oogeo.objects.line.Line([test_point, test_nearest_point], point.spatial_reference)
                if not test_line.is_perpendicular(edge, tolerance=3):
                    continue

            # if the test point is closer and out front, see if the line formed between the
            # two points are perpendicular to the edge - if so we want to ignore this
            # point for analysis
            test_line = src.oogeo.objects.line.Line([test_point, point], point.spatial_reference)
            if test_line.is_perpendicular(edge):
                analyze = False

        # if we decided not to analyze this point, continue
        if not analyze:
            continue

        # if we're here then this is a point we want to consider for transfer
        # so see if we need to snap it to a vertex or derive a point on the edge
        first_point_distance = point.distance_to(edge.first_point)
        last_point_distance = point.distance_to(edge.last_point)
        snap = first_point_distance <= tolerance or last_point_distance <= tolerance
        new_point = edge.first_point if first_point_distance <= tolerance and first_point_distance < last_point_distance else \
            edge.last_point if last_point_distance <= tolerance and last_point_distance < first_point_distance else \
            edge.nearest_point(point)

        if not point.equals(new_point):
            return_list.append({
                'point': point,
                'new_point': new_point,
                'snap': snap}
            )

    # find points in the near geometry that need to be added
    # to the target
    near_points = near_geometry.points if not isinstance(near_geometry.geometry, ShapelyPoint) else [near_geometry]
    target_boundary = target_geometry.boundary if isinstance(target_geometry.geometry, MultiPolygon) else target_geometry
    near_points = list(filter(lambda x: x.distance_to(target_boundary) <= tolerance, near_points))
    for near_point in near_points:
        found = False
        for return_object in return_list:
            if return_object['new_point'].equals(near_point):
                found = True
                break
        if not found:
            return_list.append({
                'point': None,
                'new_point': near_point,
                'snap': False}
            )

    return return_list


def insert_weld_points(geometry, weld_points):
    if isinstance(geometry.geometry, ShapelyPoint):
        return geometry
    for point in weld_points:
        if point['point']:
            geometry = geometry.replace_point(point['point'], point['new_point'])
        else:
            geometry = geometry.add_point(point['new_point'])
    return geometry


class MultipointBase(GeometryBase, abstract_base):
    """ Common base class for all geometry objects

    """

    def __init__(self, geometry, spatial_reference=None, allow_invalid=False):
        """ Initializes the geometry object.

        :param geometry: The geometry object representing the underlying geometry.
        :param spatial_reference: The spatial reference of the geometry.
        :param allow_invalid: If true, errors for invalid geometries will be ignored. Set this
            flag to true when loading datasets where you expect to encounter invalid geometries.
        """
        super(MultipointBase, self).__init__(geometry, spatial_reference, allow_invalid)

    @staticmethod
    def remove_sliver_polygons(multipolygon):
        """

        :param multipolygon:
        :return:
        """
        if not isinstance(multipolygon, MultiPolygon):
            return multipolygon
        valid_polygons = list(filter(lambda x: x.is_valid and x.area > 25, multipolygon.geoms))
        return MultiPolygon(valid_polygons) if len(valid_polygons) > 0 else None

    def intersection(self, other, exact=False, allow_invalid=False):
        """ Determines the intersecting area between this geometry and the other geometry. The method will return
        a `Polygon()` object if the geometries overlap, a `Line()` object if they only share an edge, or a `Point()`
        object if they intersect at a single point.

        :param other: The other geometry object to be evaluated.
        :param exact: If True, the geometries will be compared exactly as they are. If False, the vertices
            of the geometries will be merged prior to comparison to mitigate for tiny gaps and overlaps.
        :return: A `Polygon()` object if the geometries overlap, a `Line()` object if they only share an edge,
            or a `Point()` object if they intersect at a single point.
        """

        if exact:
            merged_self = self
            merged_other = other
        else:
            merged_geometries = MultipointBase.merge_vertices(self, other)
            merged_self = merged_geometries[0]
            merged_other = merged_geometries[1]
        # merged_self, merged_other =  if not exact else self, other
        intersection_geometries = merged_self.geometry.intersection(merged_other.geometry)

        return_geometries = []
        intersection_geometries = [geom for geom in intersection_geometries.geoms] if isinstance(intersection_geometries, GeometryCollection) \
            else [intersection_geometries] if not isinstance(intersection_geometries, list) \
            else intersection_geometries
        for geometry in intersection_geometries:
            if isinstance(geometry, ShapelyPoint):
                return_geometries.append(src.oogeo.objects.point.Point(geometry, self.spatial_reference, allow_invalid=allow_invalid))
            elif (isinstance(geometry, LineString) or isinstance(geometry, MultiLineString)) and not geometry.is_empty:
                return_geometries.append(src.oogeo.objects.line.Line(geometry, self.spatial_reference, allow_invalid=allow_invalid))
            elif isinstance(geometry, ShapelyPolygon) or isinstance(geometry, MultiPolygon):
                if geometry:
                    return_geometries.append(src.oogeo.objects.polygon.Polygon(geometry, self.spatial_reference, allow_invalid=allow_invalid))
            elif isinstance(geometry, MultiPoint):
                for point in geometry.geoms:
                    return_geometries.append(src.oogeo.objects.point.Point(point, self.spatial_reference, allow_invalid=allow_invalid))

        if len(return_geometries) == 0:
            return None
        if len(return_geometries) == 1:
            return return_geometries[0]
        return return_geometries

    def buffer(self, distance, end_style='FLAT', join_style='BEVEL', allow_invalid=False):
        """ Returns a `Polygon()` object representing the specified buffer around the geometry.

        :param distance: The distance that the buffer should expand around the geometry.
        :param end_style: The end style to apply to the buffer. Valid values are:
            'ROUND': Creates rounded ends that extend beyond the endpoints of the geometry at the specified distance.
            'FLAT': Does not extend any buffer beyond the endpoints of the geometry.
            'SQUARE': Creates square ends that extend beyond the endpoints of the geometry at the specified distance.
        :param join_style: The joint style to apply to the buffer. Valid values are:
            'ROUND': Creates rounded intersections along the edge of the buffer.
            'MITRE': Creates mitred intersections along the edge of the buffer.
            'BEVEL': Creates beveled intersections along the edge of the buffer.
        :return: A `Polygon()` object representing the specified buffer around the geometry.
        """
        if isinstance(distance, str):
            distance = string_to_units(distance, 'FEET')

        if distance == 0:
            return src.oogeo.objects.polygon.Polygon(self.geometry.buffer(distance), self.spatial_reference, allow_invalid=allow_invalid)

        buffer_polygon = src.oogeo.objects.polygon.Polygon(
            self.geometry.buffer(distance,
                resolution=100,
                cap_style=BUFFER_END_STYLES[end_style],
                join_style=BUFFER_JOIN_STYLES[join_style]),
            self.spatial_reference, allow_invalid=allow_invalid)

        if distance > 0:
            return buffer_polygon

        # for negative buffers, clip the original geometry with the
        # buffer polygon to get the real
        if buffer_polygon.is_empty:
            return self.clone()
        return self.difference(buffer_polygon, allow_invalid=allow_invalid)

    def crosses(self, other_geometry):
        """ Determines if this line object crosses the other line object.

        :param other_geometry: The other line object to be evaluated.
        :return: True if the lines cross, otherwise False.
        """
        return self.geometry.crosses(other_geometry.geometry)

    def overlaps(self, other_line):
        """ Returns True if this line overlaps the provided line, otherwise False.

        :param other_line: The line to be compared to this line.
        :return: True if this line overlaps the provided line, otherwise False.
        """
        return self.geometry.overlaps(other_line.geometry)

    def nearest_point(self, point):
        """ Returns the point along the polygon edge that is nearest to the specified point.

        :param point: The point to analyze.
        :return: `Point()` object representing a point along the edge of the polygon
            that is nearest to the specified point.
        """
        nearest_point = nearest_points(self.geometry, point.geometry)[0]
        return src.oogeo.objects.point.Point(nearest_point, self.spatial_reference)

    def nearest_vertex(self, point):
        """ Returns the vertex nearest to the specified point.

        :param point: A `Point()` object representing the point to be analyzed.
        :return: A `Point()` object representing the vertex in the line or polygon nearest to the specified point.
        """
        return_point = None
        distance = -1
        for edge_point in self.points:
            test_distance = edge_point.distance_to(point)
            if not return_point or test_distance < distance:
                return_point = edge_point
                distance = test_distance
        return return_point

    def contains(self, other_geometry, exact=False):
        """ Returns True if the polygon contains the other geometry, otherwise False.

        :param other_geometry: The geometry to be tested.
        :return: True if the polygon contains the other geometry, otherwise False.
        """
        if exact or isinstance(other_geometry, src.oogeo.objects.point.Point):
           return self.geometry.contains(other_geometry.geometry)

        if self.geometry.contains(other_geometry.geometry):
            return True

        merged_self, merged_other = self.merge_vertices(self, other_geometry)
        return merged_self.geometry.contains(merged_other.geometry)

    def contains_point(self, point, tolerance=None):
        """ Returns True if the specified point exists along the edge of the line or polygon, otherwise False.

        :param point: A `Point()` object representing the point to be examined.
        :param tolerance: A distance representing how near points can be to be considered the same point.
            If not provided, points must be exactly the same for `True` to be returned.
        :return: True if the line or polygon contains the specified point, otherwise false.
        """

        for test_point in self.points:
            if not tolerance and test_point.equals(point):
                return True
            if tolerance and test_point.distance_to(point) <= tolerance:
                return True
        return False

    @staticmethod
    def merge_vertices(first_geometry, second_geometry, tolerance=0.0000001):

        # clone the geometries so that they are not mutated
        geometry_1 = first_geometry.clone()
        geometry_2 = second_geometry.clone()
        geometry_1 = geometry_1.remove_duplicate_vertices() if isinstance(geometry_1.geometry,
                                                                          MultiPolygon) else geometry_1
        geometry_2 = geometry_2.remove_duplicate_vertices() if isinstance(geometry_2.geometry,
                                                                          MultiPolygon) else geometry_2
        # get list of vertices to evaluate for each geometry
        weld_points_1 = get_near_points(geometry_1, geometry_2, tolerance)
        weld_points_1 = list(filter(lambda x: not x['snap'], weld_points_1))
        geometry_1 = insert_weld_points(geometry_1, weld_points_1)
        weld_points_2 = get_near_points(geometry_2, geometry_1, tolerance)
        geometry_2 = insert_weld_points(geometry_2, weld_points_2)

        return geometry_1, geometry_2

    def scale(self, scale_factor):
        """ Scales the shape by multiplying each of it's vertices by the provided scale factor.

        :param scale_factor: The factor to multiple each x/y value by for each vertex in the shape.
        :return: A new geometry object with the x/y values scaled by the provided scale factor.
        """
        new_geometry = self.clone()
        for point in self.points:
            new_geometry = new_geometry.replace_point(point, src.oogeo.objects.point.Point({'x': point.x * scale_factor, 'y': point.y * scale_factor}, self.spatial_reference), self.allow_invalid)
        return new_geometry

    def transform(self, delta_x, delta_y):
        """ Moves the geometry by the specified delta values by adding the delta values to the x/y
        values of each vertex in the geometry. Providing positive delta values will generally move
        the geometry to the north and east, whereas negative values will generally move the geometry
        to the south and west.

        :param delta_x: The value to add to the x coordinate of each vertex in the geometry.
        :param delta_y: The value to add to the y coordinate of each vertex in the geometry.
        :return: A new geometry that has been transformed based on the provided inputs.
        """
        new_geometry = self.clone()
        for point in self.points:
            new_geometry = new_geometry.replace_point(point, src.oogeo.objects.point.Point({'x': point.x + delta_x, 'y': point.y + delta_y}, self.spatial_reference), allow_invalid=True)
        if self.allow_invalid or new_geometry.is_valid:
            return new_geometry
        raise ValueError('Invalid geometry detected: {}'.format(new_geometry.ewkt))



"""
Unit tests for the Workspace class. For more information see the
[API Tests](https://github.com/envelopecity/GIS/tree/dev#api-tests) section of the GIS repository.
"""


import unittest

from src.oogeo.unittests.databasetest import DatabaseTest
from src.oogeo.objects.layer import Layer
from src.oogeo.objects.table import Table
from src.oogeo.objects.workspace import Workspace


class WorkspaceClassUnitTests(DatabaseTest):

    def test_initialize_workspace(self):
        # verify we can initialize a workspace object without failure
        test_workspace = Workspace(self._databases['testing_inputs'])
        self.assertIsInstance(test_workspace, Workspace)

    def test_is_gdb(self):
        # load a gdb workspace and verify the is_gdb property
        test_workspace = Workspace(self._databases['testing_inputs'])
        self.assertTrue(test_workspace.is_gdb)

    def test_list_featureclasses(self):
        # get the list of feature classes from the test gdb (there is only 1)
        test_workspace = Workspace(self._databases['testing_inputs'])
        feature_classes = test_workspace.list_feature_classes()
        self.assertEqual(len(feature_classes), 4)
        self.assertEqual(feature_classes[0], 'adjacent_shapes')
        self.assertEqual(feature_classes[1], 'physical_blocks')
        self.assertEqual(feature_classes[2], 'snap_targets')
        self.assertEqual(feature_classes[3], 'zoning_districts')

        # test filter function
        feature_classes = test_workspace.list_feature_classes(wildcards='zon*')
        self.assertEqual(len(feature_classes), 1)
        self.assertEqual(feature_classes[0], 'zoning_districts')
        feature_classes = test_workspace.list_feature_classes(wildcards='zoning_districts*')
        self.assertEqual(len(feature_classes), 1)
        self.assertEqual(feature_classes[0], 'zoning_districts')
        feature_classes = test_workspace.list_feature_classes(wildcards='zoning_districts')
        self.assertEqual(len(feature_classes), 1)
        self.assertEqual(feature_classes[0], 'zoning_districts')
        feature_classes = test_workspace.list_feature_classes(wildcards=['zoning_districts*'])
        self.assertEqual(len(feature_classes), 1)
        self.assertEqual(feature_classes[0], 'zoning_districts')
        feature_classes = test_workspace.list_feature_classes(wildcards=['zoning_districts*', '*zoning_districts'])
        self.assertEqual(len(feature_classes), 1)
        self.assertEqual(feature_classes[0], 'zoning_districts')

    def test_create_featureclasses(self):
        # create a feature class and then delete it
        test_workspace = Workspace(self._databases['testing_inputs'])
        feature_class = test_workspace.create_feature_class('create_point_test', 'Point')
        feature_classes = test_workspace.list_feature_classes()
        self.assertIsInstance(feature_class, Layer)
        self.assertEqual(len(feature_classes), 5)
        self.assertIn('create_point_test', feature_classes)
        feature_class.delete()

        feature_class = test_workspace.create_feature_class('create_line_test', 'Polyline')
        feature_classes = test_workspace.list_feature_classes()
        self.assertIsInstance(feature_class, Layer)
        self.assertEqual(len(feature_classes), 5)
        self.assertIn('create_line_test', feature_classes)
        feature_class.delete()

        feature_class = test_workspace.create_feature_class('create_polygon_test', 'Polygon')
        feature_classes = test_workspace.list_feature_classes()
        self.assertIsInstance(feature_class, Layer)
        self.assertEqual(len(feature_classes), 5)
        self.assertIn('create_polygon_test', feature_classes)
        feature_class.delete()

        feature_classes = test_workspace.list_feature_classes()
        self.assertEqual(len(feature_classes), 4)
        self.assertNotIn('create_point_test', feature_classes)
        self.assertNotIn('create_line_test', feature_classes)
        self.assertNotIn('create_polygon_test', feature_classes)

    def test_list_tables(self):
        # get the list of tables from the test gdb (there is only `)
        test_workspace = Workspace(self._databases['testing_inputs'])
        tables = test_workspace.list_tables()
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0], 'adjacent_lots')
        tables = test_workspace.list_tables('a*')
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0], 'adjacent_lots')
        tables = test_workspace.list_tables('*s')
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0], 'adjacent_lots')
        tables = test_workspace.list_tables(['a*', '*s'])
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0], 'adjacent_lots')
        tables = test_workspace.list_tables('xy*')
        self.assertEqual(len(tables), 0)
        tables = test_workspace.list_tables('*xy')
        self.assertEqual(len(tables), 0)

    def test_load(self):
        test_workspace = Workspace(self._databases['testing_inputs'])
        test_table = test_workspace.load('adjacent_lots')
        self.assertIsInstance(test_table, Table)
        test_layer = test_workspace.load('physical_blocks')
        self.assertIsInstance(test_layer, Layer)

    def test_create_table(self):
        # create a table and then delete it
        test_workspace = Workspace(self._databases['testing_inputs'])
        table = test_workspace.create_table('create_table_test')
        tables = test_workspace.list_tables()
        self.assertIsInstance(table, Table)
        self.assertEqual(len(tables), 2)
        self.assertIn('create_table_test', tables)
        table.delete()
        tables = test_workspace.list_tables()
        self.assertEqual(len(tables), 1)
        self.assertNotIn('create_table_test', tables)

    def test_create_geodatabase(self):
        # create a new database
        dbless_connection = self._databases['testing_inputs'].copy()
        dbless_connection['database'] = None
        temp_workspace = Workspace.create_geodatabase(dbless_connection, 'test_geodatabase')
        self.assertIsInstance(temp_workspace, Workspace)
        self.assertIn('test_geodatabase', temp_workspace.list_databases())

        # copy tables and layers from one db to another
        inputs_workspace = Workspace(self._databases['testing_inputs'])
        test_table = Table('adjacent_lots', inputs_workspace)
        test_layer = Layer('adjacent_shapes', inputs_workspace)
        test_table.copy_to_new_table(test_table.table_name, temp_workspace)
        test_layer.copy_to_new_table(test_layer.table_name, temp_workspace)
        self.assertIn(test_table.table_name, temp_workspace.list_tables())
        self.assertNotIn(test_table.table_name, temp_workspace.list_feature_classes())
        self.assertIn(test_layer.table_name, temp_workspace.list_feature_classes())
        self.assertNotIn(test_layer.table_name, temp_workspace.list_tables())

        # delete the temp database
        temp_workspace.disconnect()
        Workspace.drop_geodatabase(dbless_connection, 'test_geodatabase')
        self.assertNotIn('test_geodatabase', inputs_workspace.list_databases())

    def test_schema_operations(self):
        test_workspace = Workspace(self._databases['testing_inputs'])
        self.assertFalse(test_workspace.has_schema('testing'))
        schema_names = test_workspace.list_schemas()
        self.assertEqual(len(schema_names), 1)
        self.assertIn('public', schema_names)
        new_workspace = test_workspace.create_schema('testing')
        self.assertTrue(test_workspace.has_schema('testing'))
        self.assertTrue(new_workspace.has_schema('testing'))
        self.assertEqual(test_workspace.schema, 'public')
        self.assertEqual(new_workspace.schema, 'testing')
        self.assertEqual(len(new_workspace.list_tables_and_layers()), 0)
        schema_names = test_workspace.list_schemas()
        self.assertEqual(len(schema_names), 2)
        self.assertIn('public', schema_names)
        self.assertIn('testing', schema_names)
        new_workspace.drop_schema('testing')
        self.assertFalse(test_workspace.has_schema('testing'))
        self.assertFalse(new_workspace.has_schema('testing'))
        schema_names = test_workspace.list_schemas()
        self.assertEqual(len(schema_names), 1)
        self.assertIn('public', schema_names)

if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(WorkspaceClassUnitTests)
    unittest.TextTestRunner(verbosity=2).run(suite)



BUFFER_END_STYLES = {
    'ROUND': 1,
    'FLAT': 2,
    'SQUARE': 3
}

BUFFER_JOIN_STYLES = {
    'ROUND': 1,
    'MITRE': 2,
    'BEVEL': 3
}

METADATA_TABLES = [
    'spatial_ref_sys',
    'pointcloud_formats',
]

SRIDS = {
    '2263': 102718
}
